/**
 ** mandel.c ---- Mandelbrot & Julia set rendering
 **
 ** by Richard Sanders Copyright 2008
 ** <e-mail>    - [richard at dogcreek dot ca]
 **
 **
 **  This file is part of Mandelbrot.
 **
 ** Mandelbrot is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** Mandelbrot is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with Mandelbrot.  If not, see <http://www.gnu.org/licenses/>.
 **
 **
 ** The function "CalcMandelbrot()" is based on the source code
 ** by Ashika Umanga Umagiliya and is used with permission.
 ** <e-mail>    - [umanga dot pdn at gmail dot com]
 ** The original source code for "CalcMandelBrot()" may be obtained from:
 ** http://umanga.ifreepages.com
 **
 **/
 
#include <mgrx.h>
#include <mgrxkeys.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

#define REPEAT_RATE 150   // this is the repeat rate in milliseconds
                          // for the buttons in the control panel.
                          // adjust to suit your system.
#define MAXITER 219
#define INTEGER 0
#define DOUBLE  1

struct _button
{
  int xy[4];
  int text_xy[4];
  char *mask;
  int type;
  int (*cb)(struct _button *);
  union
  {
    int *i;
    double *d;
  };
  union
  {
    int i_inc;
    int x;
    double d_inc;
  };
  union
  {
    int i_val;
    int y;
    double d_val;
  };
};


void DrawJulia(void);
void DrawMandlebrot(void);
void ControlPannel(void);
void GetFileName(char *file_name);
void SavePngFile(char *file_name);
int CancelButton(struct _button *b);
int OkButton(struct _button *b);

int file_name_len = 32;

extern GrFont helv;

// variables for the Julia & Mandelbrot sets

GrColor cl[219], buf;
double magnify;
double prev_magnify;
double aspect;
unsigned start_x = 0, start_y = 0, pixel_count;
int escape_mode = 0;    // sets the escape formula, 0 = the standard formula
int invert = 0;         // 1 = invert the color scheme
double density = 4;     // 4 to 5
int draw_what = 0;      // 0 = mandelbrot, 1 = julia

// Julia set variables
double min_r = -2.2;    // sets screen position and size on screen
double max_r = .7;      // sets screen position and size on screen
double min_i = -1.2;    // sets screen position and size on screen
double max_i = 1.2;     // sets screen position and size on screen
double m_min_r;
double m_max_r;
double m_min_i;
double m_max_i;
double r_factor;
double i_factor;

    // Change these two numbers to get different Julia sets.
    // Try numbers in [-1, 1].
    // double cr = .26;
    // double ci = .34;

double cr = .288;
double ci = .468;


// Mandelbrot set variables
double xmin = -1.4; // sets screen position and size on screen
double xmax = 1.4;  // sets screen position and size on screen
double ymin = -1.8; // sets screen position and size on screen
double ymax = 1.8;  // sets screen position and size on screen
double distort = 2; // set from 2 -> 2.2  // changes fine detail, use very small steps



struct undo_
{
  double prev_magnify;
  double magnify;
  unsigned start_x;
  unsigned start_y;
}undo[20];

int undo_count;

void UndoStack(void)
{
  if (undo_count < 19)
    ++undo_count;
  undo[undo_count].prev_magnify = prev_magnify;
  undo[undo_count].magnify = magnify;
  undo[undo_count].start_x = start_x;
  undo[undo_count].start_y = start_y;
}

void BuildColors(void)
{
  cl[0] = GrAllocColor2(0x000000);
  cl[1] = GrAllocColor2(0x330000);
  cl[2] = GrAllocColor2(0x660000);
  cl[3] = GrAllocColor2(0x990000);
  cl[4] = GrAllocColor2(0xCC0000);
  cl[5] = GrAllocColor2(0xFF0000);
  cl[6] = GrAllocColor2(0x000033);
  cl[7] = GrAllocColor2(0x330033);
  cl[8] = GrAllocColor2(0x660033);
  cl[9] = GrAllocColor2(0x990033);
  cl[10] = GrAllocColor2(0xCC0033);
  cl[11] = GrAllocColor2(0xFF0033);
  cl[12] = GrAllocColor2(0x000066);
  cl[13] = GrAllocColor2(0x330066);
  cl[14] = GrAllocColor2(0x660066);
  cl[15] = GrAllocColor2(0x990066);
  cl[16] = GrAllocColor2(0xCC0066);
  cl[17] = GrAllocColor2(0xFF0066);
  cl[18] = GrAllocColor2(0x000099);
  cl[19] = GrAllocColor2(0x330099);
  cl[20] = GrAllocColor2(0x660099);
  cl[21] = GrAllocColor2(0x990099);
  cl[22] = GrAllocColor2(0xCC0099);
  cl[23] = GrAllocColor2(0xFF0099);
  cl[24] = GrAllocColor2(0x0000CC);
  cl[25] = GrAllocColor2(0x3300CC);
  cl[26] = GrAllocColor2(0x6600CC);
  cl[27] = GrAllocColor2(0x9900CC);
  cl[28] = GrAllocColor2(0xCC00CC);
  cl[29] = GrAllocColor2(0xFF00CC);
  cl[30] = GrAllocColor2(0x0000FF);
  cl[31] = GrAllocColor2(0x3300FF);
  cl[32] = GrAllocColor2(0x6600FF);
  cl[33] = GrAllocColor2(0x9900FF);
  cl[34] = GrAllocColor2(0xCC00FF);
  cl[35] = GrAllocColor2(0xFF00FF);
  cl[36] = GrAllocColor2(0x003300);
  cl[37] = GrAllocColor2(0x333300);
  cl[38] = GrAllocColor2(0x663300);
  cl[39] = GrAllocColor2(0x993300);
  cl[40] = GrAllocColor2(0xCC3300);
  cl[41] = GrAllocColor2(0xFF3300);
  cl[42] = GrAllocColor2(0x003333);
  cl[43] = GrAllocColor2(0x333333);
  cl[44] = GrAllocColor2(0x663333);
  cl[45] = GrAllocColor2(0x993333);
  cl[46] = GrAllocColor2(0xCC3333);
  cl[47] = GrAllocColor2(0xFF3333);
  cl[48] = GrAllocColor2(0x003366);
  cl[49] = GrAllocColor2(0x333366);
  cl[50] = GrAllocColor2(0x663366);
  cl[51] = GrAllocColor2(0x993366);
  cl[52] = GrAllocColor2(0xCC3366);
  cl[53] = GrAllocColor2(0xFF3366);
  cl[54] = GrAllocColor2(0x003399);
  cl[55] = GrAllocColor2(0x333399);
  cl[56] = GrAllocColor2(0x663399);
  cl[57] = GrAllocColor2(0x993399);
  cl[58] = GrAllocColor2(0xCC3399);
  cl[59] = GrAllocColor2(0xFF3399);
  cl[60] = GrAllocColor2(0x0033CC);
  cl[61] = GrAllocColor2(0x3333CC);
  cl[62] = GrAllocColor2(0x6633CC);
  cl[63] = GrAllocColor2(0x9933CC);
  cl[64] = GrAllocColor2(0xCC33CC);
  cl[65] = GrAllocColor2(0xFF33CC);
  cl[66] = GrAllocColor2(0x0033FF);
  cl[67] = GrAllocColor2(0x3333FF);
  cl[68] = GrAllocColor2(0x6633FF);
  cl[69] = GrAllocColor2(0x9933FF);
  cl[70] = GrAllocColor2(0xCC33FF);
  cl[71] = GrAllocColor2(0xFF33FF);
  cl[72] = GrAllocColor2(0x006600);
  cl[73] = GrAllocColor2(0x336600);
  cl[74] = GrAllocColor2(0x666600);
  cl[75] = GrAllocColor2(0x996600);
  cl[76] = GrAllocColor2(0xCC6600);
  cl[77] = GrAllocColor2(0xFF6600);
  cl[79] = GrAllocColor2(0x006633);
  cl[80] = GrAllocColor2(0x336633);
  cl[81] = GrAllocColor2(0x666633);
  cl[82] = GrAllocColor2(0x996633);
  cl[83] = GrAllocColor2(0xCC6633);
  cl[84] = GrAllocColor2(0xFF6633);
  cl[85] = GrAllocColor2(0x006666);
  cl[86] = GrAllocColor2(0x336666);
  cl[87] = GrAllocColor2(0x666666);
  cl[88] = GrAllocColor2(0x996666);
  cl[89] = GrAllocColor2(0xCC6666);
  cl[90] = GrAllocColor2(0xFF6666);
  cl[91] = GrAllocColor2(0x006699);
  cl[92] = GrAllocColor2(0x336699);
  cl[93] = GrAllocColor2(0x666699);
  cl[94] = GrAllocColor2(0x996699);
  cl[95] = GrAllocColor2(0xCC6699);
  cl[96] = GrAllocColor2(0xFF6699);
  cl[97] = GrAllocColor2(0x0066CC);
  cl[98] = GrAllocColor2(0x3366CC);
  cl[99] = GrAllocColor2(0x6666CC);
  cl[101] = GrAllocColor2(0x9966CC);
  cl[102] = GrAllocColor2(0xCC66CC);
  cl[103] = GrAllocColor2(0xFF66CC);
  cl[104] = GrAllocColor2(0x0066FF);
  cl[105] = GrAllocColor2(0x3366FF);
  cl[106] = GrAllocColor2(0x6666FF);
  cl[107] = GrAllocColor2(0x9966FF);
  cl[108] = GrAllocColor2(0xCC66FF);
  cl[109] = GrAllocColor2(0xFF66FF);
  cl[110] = GrAllocColor2(0x009900);
  cl[111] = GrAllocColor2(0x339900);
  cl[112] = GrAllocColor2(0x669900);
  cl[113] = GrAllocColor2(0x999900);
  cl[114] = GrAllocColor2(0xCC9900);
  cl[115] = GrAllocColor2(0xFF9900);
  cl[116] = GrAllocColor2(0x009933);
  cl[117] = GrAllocColor2(0x339933);
  cl[118] = GrAllocColor2(0x669933);
  cl[119] = GrAllocColor2(0x999933);
  cl[120] = GrAllocColor2(0xCC9933);
  cl[121] = GrAllocColor2(0xFF9933);
  cl[122] = GrAllocColor2(0x009966);
  cl[123] = GrAllocColor2(0x339966);
  cl[124] = GrAllocColor2(0x669966);
  cl[125] = GrAllocColor2(0x999966);
  cl[126] = GrAllocColor2(0xCC9966);
  cl[127] = GrAllocColor2(0xFF9966);
  cl[128] = GrAllocColor2(0x009999);
  cl[129] = GrAllocColor2(0x339999);
  cl[130] = GrAllocColor2(0x669999);
  cl[131] = GrAllocColor2(0x999999);
  cl[132] = GrAllocColor2(0xCC9999);
  cl[133] = GrAllocColor2(0xFF9999);
  cl[134] = GrAllocColor2(0x0099CC);
  cl[135] = GrAllocColor2(0x3399CC);
  cl[136] = GrAllocColor2(0x6699CC);
  cl[137] = GrAllocColor2(0x9999CC);
  cl[138] = GrAllocColor2(0xCC99CC);
  cl[139] = GrAllocColor2(0xFF99CC);
  cl[140] = GrAllocColor2(0x0099FF);
  cl[141] = GrAllocColor2(0x3399FF);
  cl[142] = GrAllocColor2(0x6699FF);
  cl[143] = GrAllocColor2(0x9999FF);
  cl[144] = GrAllocColor2(0xCC99FF);
  cl[145] = GrAllocColor2(0xFF99FF);
  cl[146] = GrAllocColor2(0x00CC00);
  cl[147] = GrAllocColor2(0x33CC00);
  cl[148] = GrAllocColor2(0x66CC00);
  cl[149] = GrAllocColor2(0x99CC00);
  cl[150] = GrAllocColor2(0xCCCC00);
  cl[151] = GrAllocColor2(0xFFCC00);
  cl[152] = GrAllocColor2(0x00CC33);
  cl[153] = GrAllocColor2(0x33CC33);
  cl[154] = GrAllocColor2(0x66CC33);
  cl[155] = GrAllocColor2(0x99CC33);
  cl[156] = GrAllocColor2(0xCCCC33);
  cl[157] = GrAllocColor2(0xFFCC33);
  cl[158] = GrAllocColor2(0x00CC66);
  cl[159] = GrAllocColor2(0x33CC66);
  cl[160] = GrAllocColor2(0x66CC66);
  cl[161] = GrAllocColor2(0x99CC66);
  cl[162] = GrAllocColor2(0xCCCC66);
  cl[163] = GrAllocColor2(0xFFCC66);
  cl[164] = GrAllocColor2(0x00CC99);
  cl[165] = GrAllocColor2(0x33CC99);
  cl[166] = GrAllocColor2(0x66CC99);
  cl[167] = GrAllocColor2(0x99CC99);
  cl[168] = GrAllocColor2(0xCCCC99);
  cl[169] = GrAllocColor2(0xFFCC99);
  cl[170] = GrAllocColor2(0x00CCCC);
  cl[171] = GrAllocColor2(0x33CCCC);
  cl[172] = GrAllocColor2(0x66CCCC);
  cl[173] = GrAllocColor2(0x99CCCC);
  cl[174] = GrAllocColor2(0xCCCCCC);
  cl[175] = GrAllocColor2(0xFFCCCC);
  cl[176] = GrAllocColor2(0x00CCFF);
  cl[177] = GrAllocColor2(0x33CCFF);
  cl[178] = GrAllocColor2(0x66CCFF);
  cl[179] = GrAllocColor2(0x99CCFF);
  cl[180] = GrAllocColor2(0xCCCCFF);
  cl[181] = GrAllocColor2(0xFFCCFF);
  cl[182] = GrAllocColor2(0x00FF00);
  cl[183] = GrAllocColor2(0x33FF00);
  cl[184] = GrAllocColor2(0x66FF00);
  cl[185] = GrAllocColor2(0x99FF00);
  cl[186] = GrAllocColor2(0xCCFF00);
  cl[187] = GrAllocColor2(0xFFFF00);
  cl[188] = GrAllocColor2(0x00FF33);
  cl[189] = GrAllocColor2(0x33FF33);
  cl[190] = GrAllocColor2(0x66FF33);
  cl[191] = GrAllocColor2(0x99FF33);
  cl[192] = GrAllocColor2(0xCCFF33);
  cl[193] = GrAllocColor2(0xFFFF33);
  cl[194] = GrAllocColor2(0x00FF66);
  cl[195] = GrAllocColor2(0x33FF66);
  cl[196] = GrAllocColor2(0x66FF66);
  cl[197] = GrAllocColor2(0x99FF66);
  cl[198] = GrAllocColor2(0xCCFF66);
  cl[199] = GrAllocColor2(0xFFFF66);
  cl[200] = GrAllocColor2(0x00FF99);
  cl[201] = GrAllocColor2(0x33FF99);
  cl[202] = GrAllocColor2(0x66FF99);
  cl[203] = GrAllocColor2(0x99FF99);
  cl[204] = GrAllocColor2(0xCCFF99);
  cl[205] = GrAllocColor2(0xFFFF99);
  cl[206] = GrAllocColor2(0x00FFCC);
  cl[207] = GrAllocColor2(0x33FFCC);
  cl[208] = GrAllocColor2(0x66FFCC);
  cl[209] = GrAllocColor2(0x99FFCC);
  cl[210] = GrAllocColor2(0xCCFFCC);
  cl[212] = GrAllocColor2(0xFFFFCC);
  cl[213] = GrAllocColor2(0x00FFFF);
  cl[214] = GrAllocColor2(0x33FFFF);
  cl[215] = GrAllocColor2(0x66FFFF);
  cl[216] = GrAllocColor2(0x99FFFF);
  cl[217] = GrAllocColor2(0xCCFFFF);
  cl[218] = GrAllocColor2(0xFFFFFF);
  
  buf = GrAllocColor2(0xDDDEAD);
}

int main (void)
{
  char *help = " Press H for help. ";
  GrEvent ev;
  int m_x1, m_x2, m_y1, m_y2, x, y;
  GrColor cc;
  GrTextOption to;
  GrContext *ctx;

  char *st1 = "Press \"X\" or \"Q\" to Exit the program.";
  char *st2 = "Press \"Space\" bar to undo last magnify.";
  char *st3 = "Press \"F10\" to flush undo cache and redraw the Mandelbrot.";
  char *st4 = "Press \"C\" for the control pannel.";
  char *st5 = "Press \"S\" to save the screen to a *.png file.";
  char *st6 = "To select the area to magniy.";
  char *st7 = "Press and hold the left mouse button.";
  char *st8 = "Drag the mouse to the right to expand the highlight.";
  char *st9 = "Release the Left mouse button.";
  char *st10 = "Press \"M\" for the Mandelbrot, \"J\" for the Julia.";
  char *st11 = "Press any key to continue.";

  
#ifdef __WIN32__
  GrSetDriver("win32");
#elif defined __DJGPP__
  GrSetDriver("VESA");
  file_name_len = 8;
#elif defined __XWIN__
  GrSetDriver("xwin");
#elif defined __linux__
  GrSetDriver("linuxfb");
#endif
  
  GrSetMode(GR_width_height_color_graphics, 4096, 4096, 0xffff);
//  GrEventGenExpose(GR_GEN_EXPOSE_YES);

  cc = GrAllocColor(0, 255, 255);
  GrSetContext(NULL);
  GrClearContext(GrWhite());
  BuildColors();
  to.txo_chrtype = GR_BYTE_TEXT;
  to.txo_direct = GR_TEXT_RIGHT;
  to.txo_xalign = GR_ALIGN_CENTER;
  to.txo_yalign = GR_ALIGN_BOTTOM;
  to.txo_fgcolor = GrWhite();
  to.txo_bgcolor = GrBlack();
  to.txo_font = &helv;
  
  prev_magnify = magnify = 1;
  pixel_count = GrMaxX() + 1;
  aspect = (double)(GrMaxY() + 1) / (double)(GrMaxX() + 1);
  
  // aspect ratio correction for wide screens
  if (aspect != 0.75)
  {
    min_r = min_r * 0.75 / aspect;
    max_r = max_r * 0.75 / aspect;
    xmin = xmin * 0.75 / aspect;
    xmax = xmax * 0.75 / aspect;
  }
  
  undo_count = -1;

  if (!draw_what)
    DrawMandlebrot();
  else
    DrawJulia();
    
  GrDrawString(help, strlen(help), GrMaxX() >> 1, GrMaxY() - 10, &to);

  to.txo_xalign = GR_ALIGN_LEFT;
  to.txo_yalign = GR_ALIGN_TOP;
  to.txo_fgcolor = GrBlack();
  to.txo_bgcolor = GrNOCOLOR;
  
  GrEventInit();
  GrMouseSetInternalCursor(GR_MCUR_TYPE_CROSS, GrWhite() | GrXOR, GrNOCOLOR);
  GrEventGenMmove(GR_GEN_MMOVE_ALWAYS);
  GrEventFlush();
  pixel_count = GrScreenX();

  
EVENT_LOOP:
  m_x1 = m_x2 = m_y1 = m_y2 = -1;
  GrMouseDisplayCursor();
  do
  {
    GrEventWait(&ev);
    if (ev.type == GREV_MOUSE)
    {
      if (ev.p1 == GRMOUSE_LB_PRESSED)
      {
        m_x1 = ev.p2;
        if (m_x2 == -1)
          m_x2 = m_x1;
        m_y1 = ev.p3;
      }
      else if (ev.p1 == GRMOUSE_LB_RELEASED)
      {
        /* draw magnify view */
        if (ev.p2 > m_x1 )
        {
          prev_magnify = magnify;
          magnify = (double)((pixel_count) / (m_x2 - m_x1));
          start_x = (unsigned)((double)(start_x + m_x1) / prev_magnify * magnify);
          start_y = (unsigned)((double)(start_y + m_y1) / prev_magnify * magnify);
          pixel_count = (unsigned)((double)(GrScreenX()) * magnify);
          UndoStack();
          GrMouseEraseCursor();
          if (!draw_what)
            DrawMandlebrot();
          else
            DrawJulia();
        }
        goto EVENT_LOOP;
      }
    }
    else if (ev.type == GREV_MMOVE && ev.p1 == GRMOUSE_LB_PRESSED)
    {
      if (ev.p2 > m_x1)
      {
        if (m_x1 > -1 && m_x2 > m_x1)
          GrFilledBox(m_x1, m_y1, m_x2, m_y2, cc | GrXOR);
        m_x2 = ev.p2;
        if (m_x2 > m_x1)
        {
          m_y2 = m_y1 + (int)(((double)(m_x2 - m_x1) * aspect));
          GrFilledBox(m_x1, m_y1, m_x2, m_y2, cc | GrXOR);
        }
      }
    }
    if (ev.type == GREV_KEY)
    {
      switch (ev.p1)
      {
        case GrKey_s:  // save as png file
        case GrKey_S:
        {
          char file_name[file_name_len + 5];
          if (!GrPngSupport())
          {
            goto EVENT_LOOP;
            break;
          }
          if (!draw_what)
            DrawMandlebrot();
          else
            DrawJulia();
          memset(file_name, 0, file_name_len + 5);
          GetFileName(file_name);
          if (!file_name[0])
          {
            goto EVENT_LOOP;
            break;
          }
            
          strcat(file_name, ".png");
          SavePngFile(file_name);
          goto EVENT_LOOP;
          break;
        }

        case GrKey_M: // draw mandelbrot
        case GrKey_m:
          if (!draw_what)
            break;

          draw_what = 0;
          undo_count = -1;
          magnify = 1;
          start_x = start_y = 0;
          pixel_count = GrScreenX();
          undo_count = -1;
          GrMouseEraseCursor();
          DrawMandlebrot();
          goto EVENT_LOOP;
          break;

        case GrKey_J: // draw julia
        case GrKey_j:
          if (draw_what)
            break;

          draw_what = 1;
          undo_count = -1;
          magnify = 1;
          start_x = start_y = 0;
          pixel_count = GrScreenX();
          undo_count = -1;
          GrMouseEraseCursor();
          DrawJulia();
          goto EVENT_LOOP;
          break;

        case GrKey_Space: // undo, go back one screen
          if (undo_count > 0)
          {
            --undo_count;
            prev_magnify = undo[undo_count].prev_magnify;
            magnify = undo[undo_count].magnify;
            start_x = undo[undo_count].start_x;
            start_y = undo[undo_count].start_y;
            pixel_count = (unsigned)((double)GrScreenX() * magnify);
            GrMouseEraseCursor();
            if (!draw_what)
              DrawMandlebrot();
            else
              DrawJulia();
            GrEventFlush();
            goto EVENT_LOOP;
            break;
          }

        case GrKey_F10: // flush the undo cache
          magnify = 1;
          start_x = start_y = 0;
          pixel_count = GrScreenX();
          undo_count = -1;
          GrMouseEraseCursor();
          if (!draw_what)
            DrawMandlebrot();
          else
            DrawJulia();
          goto EVENT_LOOP;
          break;

        case GrKey_Q: // quit
        case GrKey_X:
        case GrKey_x:
        case GrKey_q:
          exit(0);
          break;

        case GrKey_C: // control pannel
        case GrKey_c:
          ControlPannel();
          GrEventFlush();
          goto EVENT_LOOP;
          break;
           
        case GrKey_H: // help
        case GrKey_h:
        {
          GrContext *ctx;

          ctx = GrCreateContext(361, 231, NULL, NULL);
          if (ctx == NULL)
            break;
            
          GrMouseEraseCursor();
          x = (GrScreenX() >> 1) - 170;
          y = (GrScreenY() >> 1) - 95;

          GrBitBlt(ctx, 0, 0, GrScreenContext(), x, y,
                   x + 360, y + 230, GrWRITE);

          GrFilledBox(x, y, x + 360, y + 230, GrAllocColor(230, 230, 230));
          x += 10;
          y += 10;
          GrDrawString(st1, strlen(st1), x, y, &to);
          y += 20;
          GrDrawString(st2, strlen(st2), x, y, &to);
          y += 20;
          GrDrawString(st3, strlen(st3), x, y, &to);
          y += 20;
          GrDrawString(st4, strlen(st4), x, y, &to);
          y += 20;
          GrDrawString(st5, strlen(st5), x, y, &to);
          y += 20;
          GrDrawString(st6, strlen(st6), x, y, &to);
          GrDrawString(st6, strlen(st6), x + 1, y, &to);
          y += 20;
          GrDrawString(st7, strlen(st7), x, y, &to);
          y += 20;
          GrDrawString(st8, strlen(st8), x, y, &to);
          y += 20;
          GrDrawString(st9, strlen(st9), x, y, &to);
          y += 20;
          GrDrawString(st10, strlen(st10), x, y, &to);
          y += 20;
          GrDrawString(st11, strlen(st11), x + 1, y, &to);
          GrDrawString(st11, strlen(st11 ), x, y, &to);
          
          do
            GrEventWait(&ev);
          while(ev.type != GREV_KEY);

          GrBitBlt(GrScreenContext(), (GrScreenX() >> 1) - 170,
                   (GrScreenY() >> 1) - 95, ctx, 0, 0, 360,
                   230, GrWRITE);
          GrDestroyContext(ctx);
          GrEventFlush();
          goto EVENT_LOOP;
          break;
        }
      }
    }
  }
  while(1);
}

void  CalcTxtCursor(GrTextOption *to, int *c_start, int *c_end, char *file_name,
                    int pos)
{
  if (pos)
    *c_start = GrStringWidth(file_name, pos, to);
  else
    *c_start = 0;
    
  *c_end = *c_start;
  if (file_name[pos] == 0)
    *c_end += (GrCharWidth('w', to) - 1);
  else
    *c_end += (GrCharWidth(file_name[pos] , to) - 1);
}


void GetFileName(char *file_name)
{
  GrContext *ctx, *save;
  GrTextOption to;
  int x1, x2, y1, y2, width;
  int t_xy[4];
  int pos = 0, insert = 1, c_start, c_end, cursor_top;
  GrEvent ev;
  
  struct _button button[2];
  
  to.txo_chrtype = GR_BYTE_TEXT;
  to.txo_direct = GR_TEXT_RIGHT;
  to.txo_xalign = GR_ALIGN_LEFT;
  to.txo_yalign = GR_ALIGN_TOP;
  to.txo_fgcolor = GrBlack();
  to.txo_bgcolor = GrNOCOLOR;
  to.txo_font = &helv;
  
  GrMouseEraseCursor();
  GrMouseSetInternalCursor(GR_MCUR_TYPE_ARROW, GrWhite() | GrXOR, GrNOCOLOR);
  GrMouseDisplayCursor();
  x1 = (GrScreenX() >> 1) - 160;
  x2 = (GrScreenX() >> 1) + 160;
  y1 = (GrScreenY() >> 1) - 60;
  y2 = (GrScreenY() >> 1) + 60;

  save = GrCreateContext(x2 - x1 + 1, y2 - y1 + 1, NULL, NULL);
  GrBitBlt(save, 0, 0, GrScreenContext(), x1, y1, x2, y2, GrWRITE);
  ctx = GrCreateSubContext(x1, y1, x2, y2, GrScreenContext(), NULL);
  GrSetContext(ctx);
  GrClearContext(GrAllocColor(230, 230, 230));
  GrHLine(0, GrMaxX(), 0, GrWhite());
  GrVLine(0, 0, GrMaxY(), GrWhite());
  GrHLine(0, GrMaxX(), GrMaxY(), GrBlack());
  GrVLine(GrMaxX(), 1, GrMaxY(), GrBlack());
  
  width = GrCharWidth('w', &to) * file_name_len;
  
  GrDrawString("The file will be saved in the program directory.",
               strlen("The file will be saved in the program directory."),
               10, 20, &to);
  GrDrawString("The file will be saved in the program directory.",
               strlen("The file will be saved in the program directory."),
               11, 20, &to);
               
  GrDrawString("Enter the file name below, \".png\" will be appended.",
               strlen("Enter the file name below, \".png\" will be appended."),
               10, 35, &to);
               
  button[0].xy[0] = 150;
  button[0].xy[2] = button[0].xy[0] + 38;
  button[0].xy[3] = GrMaxY() - 6;
  button[0].xy[1] = button[0].xy[3] - 15;
  button[0].cb = OkButton;
  button[0].d = 0;
  button[0].x = x1;
  button[0].y = y1;
  GrHLine(button[0].xy[0], button[0].xy[2], button[0].xy[1], GrWhite());
  GrVLine(button[0].xy[0], button[0].xy[1], button[0].xy[3], GrWhite());
  GrHLine(button[0].xy[0], button[0].xy[2], button[0].xy[3], GrBlack());
  GrVLine(button[0].xy[2], button[0].xy[1], button[0].xy[3], GrBlack());
  GrDrawString("Ok", strlen("Ok"), 160, button[0].xy[1], &to);


  button[1].xy[0] = button[0].xy[2] + 10;
  button[1].xy[1] = button[0].xy[1];
  button[1].xy[2] = button[1].xy[0] + 38;
  button[1].xy[3] = button[0].xy[3];
  button[1].cb = CancelButton;
  button[1].d = NULL;
  button[1].x = x1;
  button[1].y = y1;
  GrHLine(button[1].xy[0], button[1].xy[2], button[1].xy[1], GrWhite());
  GrVLine(button[1].xy[0], button[1].xy[1], button[1].xy[3], GrWhite());
  GrHLine(button[1].xy[0], button[1].xy[2], button[1].xy[3], GrBlack());
  GrVLine(button[1].xy[2], button[1].xy[1], button[1].xy[3], GrBlack());
  GrDrawString("Quit", strlen("Quit"), 205, button[1].xy[1], &to);
  
  t_xy[0] = 11;
  t_xy[1] = 61;
  t_xy[2] = width;
  t_xy[3] = t_xy[1] + GrCharHeight('w', &to);

  GrHLine(t_xy[0] - 1, t_xy[2] + 1, t_xy[1] - 1, GrBlack());
  GrVLine(t_xy[0] - 1, t_xy[1] - 1, t_xy[3] + 1, GrBlack());

  GrHLine(t_xy[0] - 1, t_xy[2] + 1, t_xy[3] + 1, GrWhite());
  GrVLine(t_xy[2] + 1, t_xy[1], t_xy[3] + 1, GrWhite());
  GrFilledBox(t_xy[0], t_xy[1], t_xy[2], t_xy[3], buf);

  cursor_top = t_xy[1] + ((t_xy[3] - t_xy[1]) >> 1);
  CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
  GrFilledBox(t_xy[0] + c_start, cursor_top,
              t_xy[0] + c_end, t_xy[3], buf | GrXOR );
  do
  {
GET_AGAIN:
    GrEventWait(&ev);
    if (ev.type == GREV_KEY)
    {
      if (ev.p1 == GrKey_BackSpace)
      {
        if (pos > 0)
        {
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );

          --pos;
          memcpy(file_name + pos, file_name + pos + 1, strlen(file_name + pos));
          GrFilledBox(t_xy[0], t_xy[1], t_xy[2], t_xy[3], buf);
          GrDrawString(file_name, strlen(file_name), t_xy[0], t_xy[1], &to);
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        }
        goto GET_AGAIN;
      }
      else if (ev.p1 == GrKey_Delete)
      {
        if (file_name[pos] != 0)
        {
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
          memcpy(file_name + pos, file_name + pos + 1, strlen(file_name + pos));
          GrFilledBox(t_xy[0], t_xy[1], t_xy[2], t_xy[3], buf);
          GrDrawString(file_name, strlen(file_name), t_xy[0], t_xy[1], &to);
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        }
        goto GET_AGAIN;
      }
      else if (ev.p1 == GrKey_Left)
      {
        if (pos > 0)
        {
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
          --pos;
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        }
        goto GET_AGAIN;
      }
      else if (ev.p1 == GrKey_Right)
      {
        if (file_name[pos] > 0)
        {
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
          ++pos;
          CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
          GrFilledBox(t_xy[0] + c_start, cursor_top,
                      t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        }
        goto GET_AGAIN;
      }
      else if (ev.p1 == GrKey_Insert)
      {
        CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
        GrFilledBox(t_xy[0] + c_start, cursor_top,
                    t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        insert ^= 1;
        
        if (insert)
          cursor_top = t_xy[1] + ((t_xy[3] - t_xy[1]) >> 1);
        else
          cursor_top = t_xy[1];

        GrFilledBox(t_xy[0] + c_start, cursor_top,
                    t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        goto GET_AGAIN;
      }
      
      
      else if (isalpha(ev.p1) || isdigit(ev.p1))
      {
        if (isupper(ev.p1))
          ev.p1 += 32;

        if (insert)
        {
          memmove(file_name + pos + 1, file_name + pos, strlen(file_name + pos));
          if (file_name[file_name_len] > 0)
            file_name[file_name_len] = 0;
        }
        file_name[pos] = ev.p1;
        GrFilledBox(t_xy[0], t_xy[1], t_xy[2], t_xy[3], buf);
        GrDrawString(file_name, strlen(file_name), t_xy[0], t_xy[1], &to);
        ++pos;
        if (pos >= file_name_len)
          pos = file_name_len - 1;
        CalcTxtCursor(&to, &c_start, &c_end, file_name, pos);
        GrFilledBox(t_xy[0] + c_start, cursor_top,
                    t_xy[0] + c_end, t_xy[3], buf | GrXOR );
        goto GET_AGAIN;
      }
      else if (ev.p1 == GrKey_Return)
        break;
    }
    else if (ev.type == GREV_MOUSE && ev.p1 == GRMOUSE_LB_PRESSED)
    {
      ev.p2 -= x1;
      ev.p3 -= y1;
      if (ev.p2 >= button[0].xy[0] && ev.p2 <= button[0].xy[2] &&
          ev.p3 >= button[0].xy[1] && ev.p3 <= button[0].xy[3])
      {
        if (button[0].cb(&button[0]))
          break;
      }
      else if (ev.p2 >= button[1].xy[0] && ev.p2 <= button[1].xy[2] &&
          ev.p3 >= button[1].xy[1] && ev.p3 <= button[1].xy[3])
      {
        if (button[1].cb(&button[1]))
        {
          file_name[0] = 0;
          break;
        }
      }
    }
  }
  while(1);
  
  GrMouseEraseCursor();
  GrMouseSetInternalCursor(GR_MCUR_TYPE_CROSS, GrWhite() | GrXOR, GrNOCOLOR);
  GrBitBlt(GrScreenContext(), x1, y1, save, 0, 0, GrMaxX(), GrMaxY(), GrWRITE);
  GrSetContext(GrScreenContext());
  GrDestroyContext(save);
  GrDestroyContext(ctx);
}

void SavePngFile(char *file_name)
{
  GrSaveContextToPng(GrScreenContext(), file_name);
  GrMouseDisplayCursor();
}

int CalcEscape(double r, double i, double escape_value)
{
  int escape_result = 0;
  switch (escape_mode)
  {
    case 0:
      if (r * r + i * i > escape_value)
        escape_result = 1;
      break;

    case 1:
      if (r * r > escape_value && i * 1 > escape_value)
        escape_result = 1;
      break;

    case 2:
      if (r > escape_value || i > escape_value)
        escape_result = 1;
      break;

    case 3:
      if ((r + i) * (r + i) > escape_value)
        escape_result = 1;
      break;

    case 4:
      if ((fabs(r) + fabs(i)) * (fabs(r) + fabs(i)) > escape_value)
        escape_result = 1;
      break;

    case 5:
      if (i * i > escape_value)
        escape_result = 1;
      break;

    case 6:
      if (r * r > escape_value)
        escape_result = 1;
      break;

    case 7:
      if (fabs(2 * r * i) > escape_value)
        escape_result = 1;
      break;

    case 8:
      if (r * r > escape_value || i * i > escape_value)
        escape_result = 1;
      break;

    case 9:
      if (fabs(r) > escape_value || fabs(i) > escape_value)
        escape_result = 1;
      break;
  }
  return escape_result;
}


int CalcJulia(double r, double i)
{
  double newr, newi;
  int loop;
  
  for(loop = 0; loop < 219; ++loop)
  {
    newr = r * r - i * i + cr;
    newi = r * i + r * i + ci;
    r = newr;
    i = newi;

    if (CalcEscape(r, i, density))
      return loop;
  }
  return loop;
}

void DrawJulia(void)
{
  double r;
  double i;
  double rinc;
  double iinc;
  double j_min_x = xmin;
  double j_max_x = xmax;
  double j_min_y = ymin;
  double j_max_y = ymax;
  int x, y, point;
  GrColor *scanline;
  
  // scaling for magnify and position
  if (magnify > 1)
  {
    double inc;

    inc = (j_max_x - j_min_x) / pixel_count;
    if (start_x)
      j_min_x += (double)start_x * inc;
    j_max_x = j_min_x + inc * (double)GrMaxX();

    inc = (j_max_y - j_min_y) / (pixel_count * aspect);
    if (start_y)
      j_min_y += (double)start_y * inc;
    j_max_y = j_min_y + inc * (double)GrMaxY();
  }

  rinc = (j_max_x - j_min_x) / (GrMaxX() + 1);
  iinc = (j_max_y - j_min_y) / (GrMaxY() + 1);
  i = j_min_y;
  for(y = 0; y <= GrMaxY(); y++)
  {
    scanline = (GrColor *)GrGetScanline (0, GrMaxX(), y);
    r = j_min_x;
    for(x = 0; x <= GrMaxX(); x++)
    {
      point = CalcJulia(r, i);
      if (invert)
        point = 219 - point;
      scanline[x] = cl[point];
      r += rinc;
    }
    i += iinc;
    GrPutScanline (0, GrMaxX(), y, scanline, GrWRITE);
  }
	GrMouseWarp(GrMaxX() >> 1, GrMaxY() >> 1);
}


int CalcMandelbrot(int x, int y)
{
  double c_r = m_min_r + x * r_factor;   //real part of C
  double c_i = m_min_i + y * i_factor; //imaginary part of C
  double z_r = c_r, z_i = c_i; //Initially Z(0) is C
  int loop;

  for (loop = 0; loop < MAXITER; loop++)
  {
    double z_r2 = z_r * z_r, z_i2 = z_i * z_i;
    if (CalcEscape(z_r, z_i, density))
      return loop;
      
    z_i = distort * z_r * z_i + c_i;
    z_r = z_r2 - z_i2 + c_r;
  }
  return loop;
}

void DrawMandlebrot(void)
{
  int x, y, point;
  GrColor *scanline;

  m_min_r = min_r;
  m_max_r = max_r;
  m_min_i = min_i;
  m_max_i = max_i;
  // scaling for magnify and position
  if (magnify > 1)
  {
    double inc;

    inc = (m_max_r - m_min_r) / pixel_count;
    if (start_x)
      m_min_r += (double)start_x * inc;
    m_max_r = m_min_r + inc * (double)GrMaxX();

    inc = (m_max_i - m_min_i) / (pixel_count * aspect);
    if (start_y)
      m_min_i += (double)start_y * inc;
    m_max_i = m_min_i + inc * (double)GrMaxY();
  }

  r_factor = (m_max_r - m_min_r) / (GrMaxX() + 1);
  i_factor = (m_max_i - m_min_i) / (GrMaxY() + 1);
  
  for(y = 0; y <= GrMaxY(); ++y)
  {
    scanline = (GrColor *)GrGetScanline (0, GrMaxX(), y);
    for(x = 0; x <= GrMaxX(); ++x)
    {
      point = CalcMandelbrot (x, y);
      if (invert)
        point = 219 - point;
      scanline[x] = cl[point];
    }
    GrPutScanline (0, GrMaxX(), y, scanline, GrWRITE);
  }
  GrMouseWarp(GrMaxX() >> 1, GrMaxY() >> 1);
}

void UpButton(struct _button *b)
{
  GrFilledBox(b->xy[0], b->xy[1], b->xy[2],b->xy[3],
              GrAllocColor(230, 230, 230));
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrWhite());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrWhite());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrBlack());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrBlack());

  GrPlot(b->xy[0] + 5, b->xy[1] + 2, GrBlack());
  GrHLine(b->xy[0] + 4, b->xy[0] + 6, b->xy[1] + 3, GrBlack());
}

void DownButton(struct _button *b)
{
  GrFilledBox(b->xy[0], b->xy[1], b->xy[2],b->xy[3],
              GrAllocColor(230, 230, 230));
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrWhite());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrWhite());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrBlack());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrBlack());

  GrPlot(b->xy[0] + 5, b->xy[1] + 3, GrBlack());
  GrHLine(b->xy[0] + 4, b->xy[0] + 6, b->xy[1] + 2, GrBlack());
}

void UpButtonPressed(struct _button *b)
{
  GrFilledBox(b->xy[0], b->xy[1], b->xy[2],b->xy[3],
              GrAllocColor(230, 230, 230));
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrBlack());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrBlack());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrWhite());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrWhite());

  GrPlot(b->xy[0] + 6, b->xy[1] + 3, GrBlack());
  GrHLine(b->xy[0] + 5, b->xy[0] + 7, b->xy[1] + 4, GrBlack());
}

void DownButtonPressed(struct _button *b)
{
  GrFilledBox(b->xy[0], b->xy[1], b->xy[2],b->xy[3],
              GrAllocColor(230, 230, 230));
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrBlack());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrBlack());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrWhite());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrWhite());

  GrPlot(b->xy[0] + 6, b->xy[1] + 4, GrBlack());
  GrHLine(b->xy[0] + 5, b->xy[0] + 7, b->xy[1] + 3, GrBlack());
}

void ButtonUp(struct _button *b)
{
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrWhite());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrWhite());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrBlack());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrBlack());
}

void ButtonDown(struct _button *b)
{
  GrHLine(b->xy[0], b->xy[2], b->xy[1], GrBlack());
  GrVLine(b->xy[0], b->xy[1], b->xy[3], GrBlack());
  GrHLine(b->xy[0] + 1, b->xy[2], b->xy[3], GrWhite());
  GrVLine(b->xy[2], b->xy[1] + 1, b->xy[3], GrWhite());
}

void AdjustValue(struct _button *b)
{
  char str[20];
  GrTextOption to;
  
  to.txo_chrtype = GR_BYTE_TEXT;
  to.txo_direct = GR_TEXT_RIGHT;
  to.txo_xalign = GR_ALIGN_LEFT;
  to.txo_yalign = GR_ALIGN_TOP;
  to.txo_fgcolor = GrBlack();
  to.txo_bgcolor = GrNOCOLOR;
  to.txo_font = &helv;

  if (b->type == INTEGER)
  {
    *b->i += b->i_inc;
    if (b->i_inc < 0)
    {
      if (*b->i < b->i_val)
        *b->i = b->i_val;
    }
    else
    {
      if (*b->i > b->i_val)
        *b->i = b->i_val;
    }
  }
  else
  {
    *b->d += b->d_inc;
    if (b->d_inc < 0)
    {
      if (*b->d < b->d_val)
        *b->d = b->d_val;
    }
    else
    {
      if (*b->d > b->d_val)
        *b->d = b->d_val;
    }
  }
  if (b->mask[0] == 'N')
  {
    if (*b->i)
      strcpy(str, "Yes");
    else
      strcpy(str, "No");
  }
  else if (b->type == INTEGER)
    sprintf(str, b->mask, *b->i);
  else
    sprintf(str, b->mask, *b->d);
    
  GrFilledBox(b->text_xy[0] + 1, b->text_xy[1] + 1,
              b->text_xy[2] - 1, b->text_xy[3] - 1,
              GrWhite());
  GrDrawString(str, strlen(str), b->text_xy[0] + 3, b->text_xy[1], &to);
}


int Spinner(struct _button *b)
{
  int direction, on_button = 1;
  long t;
  GrEvent ev;

  t = GrMsecTime();
  
  if (b->type == INTEGER)
  {
    if (b->i_inc < 0)
      direction = -1;
    else
      direction = 1;
  }
  else
  {
    if (b->d_inc < 0)
      direction = -1;
    else
      direction = 1;
  }
  
  if (direction < 0)
    DownButtonPressed(b);
  else
    UpButtonPressed(b);
  AdjustValue(b);
  goto REPEAT;
  
  do
  {
    GrEventWait(&ev);
    if (ev.type == GREV_MOUSE || ev.type == GREV_MMOVE)
    {
      if (ev.p1 == GRMOUSE_LB_RELEASED)
      {
        if (direction < 0)
          DownButton(b);
        else
          UpButton(b);
        return 0;
      }
      
      ev.p2 -= (GrScreenX() >> 1) - 90;
      ev.p3 -= (GrScreenY() >> 1) - 115;
      if (ev.p2 < b->xy[0] || ev.p2 > b->xy[2] ||
          ev.p3 < b->xy[1] || ev.p3 > b->xy[3] )
      { // mouse off button
        if (on_button)
        {
          on_button = 0;
          if (direction < 0)
            DownButton(b);
          else
            UpButton(b);
        }
      }
      else if (!on_button)
      { // mouse moved back over button
        on_button = 1;
        if (direction < 0)
          DownButtonPressed(b);
        else
          UpButtonPressed(b);
REPEAT:
        while (!GrEventCheck())
        {
          if (GrMsecTime() - t >= REPEAT_RATE)
          { // repeat delay expired
            t = GrMsecTime();
            AdjustValue(b);
          }
        }
      }
      else
        goto REPEAT;
    }
  }
  while (1);
  return 0;
}

int OkButton(struct _button *b)
{
  int pressed = 1;
  GrEvent ev;
  
  ButtonDown(b);
  do
  {
    GrEventWait(&ev);
    if (ev.type == GREV_MOUSE || ev.type == GREV_MMOVE)
    {
      ev.p2 -= (GrScreenX() >> 1) - b->x;
      ev.p3 -= (GrScreenY() >> 1) - b->y;

      if (ev.p1 == GRMOUSE_LB_RELEASED)
      {
        ButtonUp(b);
        if (ev.p2 < b->xy[0] || ev.p2 > b->xy[2] ||
            ev.p3 < b->xy[1] || ev.p3 > b->xy[3] )
          return 1;
        else
          return 0;
      }

      if (ev.p2 < b->xy[0] || ev.p2 > b->xy[2] ||
          ev.p3 < b->xy[1] || ev.p3 > b->xy[3] )
      {
        ButtonUp(b);
        pressed = 0;
      }
      else if (!pressed)
      {
        pressed = 1;
        ButtonDown(b);
      }
    }
  }
  while(1);
  return 0;
}

int CancelButton(struct _button *b)
{
  int pressed = 1;
  GrEvent ev;

  ButtonDown(b);
  do
  {
    GrEventWait(&ev);
    if (ev.type == GREV_MOUSE || ev.type == GREV_MMOVE)
    {
      ev.p2 -= (GrScreenX() >> 1) - b->x;
      ev.p3 -= (GrScreenY() >> 1) - b->y;
      if (ev.p1 == GRMOUSE_LB_RELEASED)
      {
        ButtonUp(b);
        if (ev.p2 < b->xy[0] || ev.p2 > b->xy[2] ||
            ev.p3 < b->xy[1] || ev.p3 > b->xy[3] )
          return 1;
        else
          return 0;
      }

      if (ev.p2 < b->xy[0] || ev.p2 > b->xy[2] ||
          ev.p3 < b->xy[1] || ev.p3 > b->xy[3] )
      {
        ButtonUp(b);
        pressed = 0;
      }
      else if (!pressed)
      {
        pressed = 1;
        ButtonDown(b);
      }
    }
  }
  while(1);
  return 0;
}

void ControlPannel(void)
{
  int x1, x2, y1, y2, loop, return_val = 0;
  GrEvent ev;
  GrContext *ctx, *save;
  GrTextOption to;
  char str[20];
  int dup_invert = invert;
  int dup_escape_mode = escape_mode;
  double dup_density = density;
  double dup_distort = distort;
  double dup_cr = cr;
  double dup_ci = ci;
  struct _button button[14];
  
  
  GrMouseEraseCursor();
  GrMouseSetInternalCursor(GR_MCUR_TYPE_ARROW, GrWhite() | GrXOR, GrNOCOLOR);
  x1 = (GrScreenX() >> 1) - 90;
  x2 = (GrScreenX() >> 1) + 90;
  y1 = (GrScreenY() >> 1) - 115;
  y2 = (GrScreenY() >> 1) + 115;
  save = GrCreateContext(x2 - x1 + 1, y2 - y1 + 1, NULL, NULL);
  GrBitBlt(save, 0, 0, GrScreenContext(), x1, y1, x2, y2, GrWRITE);
  ctx = GrCreateSubContext(x1, y1, x2, y2, GrScreenContext(), NULL);
  GrSetContext(ctx);
  GrClearContext(GrAllocColor(230, 230, 230));

  GrHLine(0, GrMaxX(), 0, GrWhite());
  GrVLine(0, 0, GrMaxY(), GrWhite());
  GrHLine(0, GrMaxX(), GrMaxY(), GrBlack());
  GrVLine(GrMaxX(), 1, GrMaxY(), GrBlack());

  to.txo_chrtype = GR_BYTE_TEXT;
  to.txo_direct = GR_TEXT_RIGHT;
  to.txo_xalign = GR_ALIGN_LEFT;
  to.txo_yalign = GR_ALIGN_TOP;
  to.txo_fgcolor = GrBlack();
  to.txo_bgcolor = GrNOCOLOR;
  to.txo_font = &helv;

  GrDrawString("Escape mode", strlen("Escape mode"), 10, 10, &to);
  
  GrFilledBox(100, 10, 113, 23, GrWhite());
  GrHLine(100, 113, 10, cl[131]);
  GrVLine(100, 10, 23, cl[131]);
  GrHLine(101, 113, 23, cl[174]);
  GrVLine(113, 11, 23, cl[174]);
  sprintf(str, "%d", dup_escape_mode);
  GrDrawString(str, strlen(str), 103, 10, &to);

  button[0].xy[0] = 115;
  button[0].xy[1] = 10;
  button[0].xy[2] = button[0].xy[0] + 11;
  button[0].xy[3] = button[0].xy[1] + 6;
  button[0].cb = Spinner;
  button[0].i = &dup_escape_mode;
  button[0].i_inc = 1;
  button[0].i_val = 9;
  button[0].text_xy[0] = 100;
  button[0].text_xy[1] = 10;
  button[0].text_xy[2] = 113;
  button[0].text_xy[3] = 23;
  button[0].mask = "%d";
  button[0].type = INTEGER;
  UpButton(&button[0]);
  
  button[1].xy[0] = 115;
  button[1].xy[1] = 17;
  button[1].xy[2] = button[1].xy[0] + 11;
  button[1].xy[3] = button[1].xy[1] + 6;
  button[1].cb = Spinner;
  button[1].i = &dup_escape_mode;
  button[1].i_inc = -1;
  button[1].i_val = 0;
  button[1].text_xy[0] = 100;
  button[1].text_xy[1] = 10;
  button[1].text_xy[2] = 113;
  button[1].text_xy[3] = 23;
  button[1].mask = "%d";
  button[1].type = INTEGER;
  DownButton(&button[1]);
  
  GrDrawString("Escape value", strlen("Escape value"), 10, 30, &to);
  GrFilledBox(100, 30, 133, 43, GrWhite());
  GrHLine(100, 133, 30, cl[131]);
  GrVLine(100, 30, 43, cl[131]);
  GrHLine(101, 133, 43, cl[174]);
  GrVLine(133, 31, 43, cl[174]);
  sprintf(str, "%3.2f", dup_density);
  GrDrawString(str, strlen(str), 103, 30, &to);

  button[2].xy[0] = 135;
  button[2].xy[1] = 30;
  button[2].xy[2] = button[2].xy[0] + 11;
  button[2].xy[3] = button[2].xy[1] + 6;
  button[2].cb = Spinner;
  button[2].d = &dup_density;
  button[2].d_inc = .01;
  button[2].d_val = 5;
  button[2].text_xy[0] = 100;
  button[2].text_xy[1] = 30;
  button[2].text_xy[2] = 133;
  button[2].text_xy[3] = 43;
  button[2].mask = "%3.2f";
  button[2].type = DOUBLE;
  UpButton(&button[2]);
  
  button[3].xy[0] = 135;
  button[3].xy[1] = 37;
  button[3].xy[2] = button[3].xy[0] + 11;
  button[3].xy[3] = button[3].xy[1] + 6;
  button[3].cb = Spinner;
  button[3].d = &dup_density;
  button[3].d_inc = -.01;
  button[3].d_val = 4;
  button[3].text_xy[0] = 100;
  button[3].text_xy[1] = 30;
  button[3].text_xy[2] = 133;
  button[3].text_xy[3] = 43;
  button[3].mask = "%3.2f";
  button[3].type = DOUBLE;
  DownButton(&button[3]);
  
  GrDrawString("Invert color", strlen("Invert color"), 10, 50, &to);
  if (dup_invert)
    strcpy(str, "Yes");
  else
    strcpy(str, "No");
  GrFilledBox(100, 50, 133, 63, GrWhite());
  GrHLine(100, 133, 50, cl[131]);
  GrVLine(100, 50, 63, cl[131]);
  GrHLine(101, 133, 63, cl[174]);
  GrVLine(133, 51, 63, cl[174]);
  GrDrawString(str, strlen(str), 103, 50, &to);

  button[4].xy[0] = 135;
  button[4].xy[1] = 50;
  button[4].xy[2] = button[4].xy[0] + 11;
  button[4].xy[3] = button[4].xy[1] + 6;
  button[4].cb = Spinner;
  button[4].i = &dup_invert;
  button[4].i_inc = 1;
  button[4].i_val = 1;
  button[4].text_xy[0] = 100;
  button[4].text_xy[1] = 50;
  button[4].text_xy[2] = 133;
  button[4].text_xy[3] = 63;
  button[4].mask = "N";
  button[4].type = INTEGER;
  UpButton(&button[4]);

  button[5].xy[0] = 135;
  button[5].xy[1] = 57;
  button[5].xy[2] = button[5].xy[0] + 11;
  button[5].xy[3] = button[5].xy[1] + 6;
  button[5].cb = Spinner;
  button[5].i = &dup_invert;
  button[5].i_inc = -1;
  button[5].i_val = 0;
  button[5].text_xy[0] = 100;
  button[5].text_xy[1] = 50;
  button[5].text_xy[2] = 133;
  button[5].text_xy[3] = 63;
  button[5].mask = "N";
  button[5].type = INTEGER;
  DownButton(&button[5]);
  
  GrDrawString("Julia Set Only", strlen("Julia Set Only"), 30, 80, &to);
  GrDrawString("Julia Set Only", strlen("Julia Set Only"), 31, 80, &to);
  
  GrDrawString("C real", strlen("C real"), 10, 100, &to);
  GrFilledBox(100, 100, 153, 113, GrWhite());
  GrHLine(100, 153, 100, cl[131]);
  GrVLine(100, 100, 113, cl[131]);
  GrHLine(101, 153, 113, cl[174]);
  GrVLine(153, 101, 113, cl[174]);
  
  sprintf(str, "%4.3f", dup_cr);
  GrDrawString(str, strlen(str), 103, 100, &to);

  button[6].xy[0] = 155;
  button[6].xy[1] = 100;
  button[6].xy[2] = button[6].xy[0] + 11;
  button[6].xy[3] = button[6].xy[1] + 6;
  button[6].cb = Spinner;
  button[6].d = &dup_cr;
  button[6].d_inc = .001;
  button[6].d_val = 1;
  button[6].text_xy[0] = 100;
  button[6].text_xy[1] = 100;
  button[6].text_xy[2] = 153;
  button[6].text_xy[3] = 113;
  button[6].mask = "%4.3f";
  button[6].type = DOUBLE;
  UpButton(&button[6]);

  button[7].xy[0] = 155;
  button[7].xy[1] = 107;
  button[7].xy[2] = button[7].xy[0] + 11;
  button[7].xy[3] = button[7].xy[1] + 6;
  button[7].cb = Spinner;
  button[7].d = &dup_cr;
  button[7].d_inc = -.001;
  button[7].d_val = -1;
  button[7].text_xy[0] = 100;
  button[7].text_xy[1] = 100;
  button[7].text_xy[2] = 153;
  button[7].text_xy[3] = 113;
  button[7].mask = "%4.3f";
  button[7].type = DOUBLE;
  DownButton(&button[7]);
  
  GrDrawString("C imaginary", strlen("C imaginary"), 10, 120, &to);
  GrFilledBox(100, 120, 153, 133, GrWhite());
  GrHLine(100, 153, 120, cl[131]);
  GrVLine(100, 120, 133, cl[131]);
  GrHLine(101, 153, 133, cl[174]);
  GrVLine(153, 121, 133, cl[174]);

  sprintf(str, "%4.3f", dup_ci);
  GrDrawString(str, strlen(str), 103, 120, &to);

  button[8].xy[0] = 155;
  button[8].xy[1] = 120;
  button[8].xy[2] = button[8].xy[0] + 11;
  button[8].xy[3] = button[8].xy[1] + 6;
  button[8].cb = Spinner;
  button[8].d = &dup_ci;
  button[8].d_inc = .001;
  button[8].d_val = 1;
  button[8].text_xy[0] = 100;
  button[8].text_xy[1] = 120;
  button[8].text_xy[2] = 153;
  button[8].text_xy[3] = 133;
  button[8].mask = "%4.3f";
  button[8].type = DOUBLE;
  UpButton(&button[8]);
  
  button[9].xy[0] = 155;
  button[9].xy[1] = 127;
  button[9].xy[2] = button[9].xy[0] + 11;
  button[9].xy[3] = button[9].xy[1] + 6;
  button[9].cb = Spinner;
  button[9].d = &dup_ci;
  button[9].d_inc = -.001;
  button[9].d_val = -1;
  button[9].text_xy[0] = 100;
  button[9].text_xy[1] = 120;
  button[9].text_xy[2] = 153;
  button[9].text_xy[3] = 133;
  button[9].mask = "%4.3f";
  button[9].type = DOUBLE;
  DownButton(&button[9]);
  
  GrDrawString("Mandelbrot only", strlen("Mandelbrot only"), 30, 150, &to);
  GrDrawString("Mandelbrot only", strlen("Mandelbrot only"), 31, 150, &to);

  GrDrawString("Distort", strlen("Distort"), 10, 170, &to);
  GrFilledBox(100, 170, 130, 183, GrWhite());
  GrHLine(100, 130, 170, cl[131]);
  GrVLine(100, 170, 183, cl[131]);
  GrHLine(101, 130, 183, cl[174]);
  GrVLine(130, 171, 183, cl[174]);
  sprintf(str, "%3.2f", dup_distort);
  GrDrawString(str, strlen(str), 103, 170, &to);

  button[10].xy[0] = 132;
  button[10].xy[1] = 170;
  button[10].xy[2] = button[10].xy[0] + 11;
  button[10].xy[3] = button[10].xy[1] + 6;
  button[10].cb = Spinner;
  button[10].d = &dup_distort;
  button[10].d_inc = .01;
  button[10].d_val = 2.2;
  button[10].text_xy[0] = 100;
  button[10].text_xy[1] = 170;
  button[10].text_xy[2] = 130;
  button[10].text_xy[3] = 183;
  button[10].mask = "%3.2f";
  button[10].type = DOUBLE;
  UpButton(&button[10]);
  
  button[11].xy[0] = 132;
  button[11].xy[1] = 177;
  button[11].xy[2] = button[11].xy[0] + 11;
  button[11].xy[3] = button[11].xy[1] + 6;
  button[11].cb = Spinner;
  button[11].d = &dup_distort;
  button[11].d_inc = -.01;
  button[11].d_val = 2;
  button[11].text_xy[0] = 100;
  button[11].text_xy[1] = 170;
  button[11].text_xy[2] = 130;
  button[11].text_xy[3] = 183;
  button[11].mask = "%3.2f";
  button[11].type = DOUBLE;
  DownButton(&button[11]);
  
  GrDrawString("Ok", strlen("Ok"), 45, 200, &to);
  GrHLine(20, 85, 198, GrWhite());
  GrVLine(20, 198, 218, GrWhite());
  GrHLine(20, 85, 218, GrBlack());
  GrVLine(85, 198, 218, GrBlack());
  button[12].xy[0] = 20;
  button[12].xy[1] = 198;
  button[12].xy[2] = 85;
  button[12].xy[3] = 218;
  button[12].cb = OkButton;
  button[12].d = 0;
  button[12].x = x1;
  button[12].y = y1;
  
  GrDrawString("Quit", strlen("Quit"), 118, 200, &to);
  GrHLine(95, 160, 198, GrWhite());
  GrVLine(95, 198, 218, GrWhite());
  GrHLine(95, 160, 218, GrBlack());
  GrVLine(160, 198, 218, GrBlack());
  button[13].xy[0] = 95;
  button[13].xy[1] = 198;
  button[13].xy[2] = 160;
  button[13].xy[3] = 218;
  button[13].cb = CancelButton;
  button[13].d = NULL;
  button[13].x = x1;
  button[13].y = y1;
  
  GrMouseDisplayCursor();
  
  do
  {
    GrEventWait(&ev);
    
    if (ev.type == GREV_MOUSE)
    {
      ev.p2 -= x1;
      ev.p3 -= y1;
      for(loop = 0; loop < 14; ++loop)
      {
        if (ev.p2 >= button[loop].xy[0] &&
            ev.p2 <= button[loop].xy[2] &&
            ev.p3 >= button[loop].xy[1] &&
            ev.p3 <= button[loop].xy[3])
        {
          return_val = button[loop].cb(&button[loop]);
          break;
        }
      }
    }
    if ((loop == 12 || loop == 13) && return_val)
    {
      if (loop == 12)
      {
        if (invert != dup_invert)

          invert = dup_invert;
        if (escape_mode != dup_escape_mode)

          escape_mode = dup_escape_mode;
        if (density != dup_density)

          density = dup_density;
        if (distort != dup_distort)

          distort = dup_distort;
        if (cr != dup_cr)
          cr = dup_cr;
        if (ci != dup_ci)
          ci = dup_ci;
      }
      GrMouseEraseCursor();
      GrMouseSetInternalCursor(GR_MCUR_TYPE_CROSS, GrWhite() | GrXOR, GrNOCOLOR);
      GrBitBlt(GrScreenContext(), x1, y1, save, 0, 0, GrMaxX(), GrMaxY(), GrWRITE);
      GrSetContext(GrScreenContext());
      GrDestroyContext(save);
      GrDestroyContext(ctx);
      
      if (!draw_what)
        DrawMandlebrot();
      else
        DrawJulia();
      return;
    }
  }
  while(1);
}
